class ProgressBar:
    def __init__(self, nb_iters):
        self._nb_iters = nb_iters
        self._iter_i = 0
        self._total_loss = 0
        self._total_metrics = 0

    def print_prog_bar(self, loss, metrics, lr_list, length=30):
        self._iter_i += 1
        prob_norm = length / self._nb_iters
        done = '=' * int(self._iter_i * prob_norm)
        todo = ' ' * int((self._nb_iters - self._iter_i) * prob_norm)
        cout = f"[{done + todo}] {self._iter_i}/{self._nb_iters}"
        self._total_loss += loss
        cout += f" loss: {(self._total_loss / self._iter_i):.4g}"
        if metrics:
            self._total_metrics += metrics
            cout += f" dice: {(self._total_metrics / self._iter_i):.4g}"
        if lr_list:
            cout += " lr: " + ",".join([f"{lr:.3g}" for lr in lr_list])
        print("\r"+cout, end="")

    def get_latest_metrics(self):
        return self._total_metrics / self._iter_i



from .get_annotation import GetAnnotation
from .get_dataset import GetDatasetSgm
from .get_optimization import GetOptimization

from . import metrics
from . import logger
from . import readers

__all__ = ['GetAnnotation', 'GetDatasetSgm', 'GetOptimization',
           'metrics', 'logger', 'readers']



def setup_for_distributed(is_master):
    """
    This function disables printing when not in master process
    """
    import builtins as __builtin__
    builtin_print = __builtin__.print

    def print(*args, **kwargs):
        force = kwargs.pop('force', False)
        if is_master or force:
            builtin_print(*args, **kwargs)

    __builtin__.print = print



import torch


class SegMetrics:
    def __init__(self):
        self.total_confusion = None

    @torch.no_grad()
    def calc_confusion(
            self, output: torch.LongTensor, target: torch.LongTensor,
            from_logits=False, threshold=None):
        if from_logits:
            output = torch.nn.functional.logsigmoid(output).exp()
        if threshold is not None:
            output = torch.where(output >= threshold, 1, 0)
            target = torch.where(target >= threshold, 1, 0)

        batch_size, num_classes, *dims = target.shape
        output = output.view(batch_size, num_classes, -1)
        target = target.view(batch_size, num_classes, -1)

        tp = (output * target).sum(2).sum(0, keepdim=True)
        fp = output.sum(2).sum(0, keepdim=True) - tp
        fn = target.sum(2).sum(0, keepdim=True) - tp
        return torch.cat((tp, fp, fn))

    def compute_metric(self, confusion, metric_fn, class_weights=1.0):
        if self.total_confusion is None:
            self.total_confusion = confusion
        else:
            self.total_confusion += confusion
        tp, fp, fn = self.total_confusion
        class_weights = torch.tensor(class_weights).to(tp.device)
        class_weights = class_weights / class_weights.sum()
        score = metric_fn(tp, fp, fn)
        score = self._handle_zero_division(score)
        score = (score * class_weights).mean()
        return score

    def _handle_zero_division(self, x):
        nans = torch.isnan(x)
        value = torch.tensor(0, dtype=x.dtype).to(x.device)
        x = torch.where(nans, value, x)
        return x

    @staticmethod
    def dice(tp, fp, fn, beta=1):
        beta_tp = (1 + beta ** 2) * tp
        beta_fn = (beta ** 2) * fn
        score = beta_tp / (beta_tp + beta_fn + fp)
        return score



from typing import List, Any
from .readers import cross_val
import dataclasses


@dataclasses.dataclass
class GetAnnotation:
    target: str

    # for train annotation
    train_dirs: List[str] = dataclasses.field(default_factory=list)
    get_train_fn: Any = None

    # for val annotation
    val_dirs: List[str] = dataclasses.field(default_factory=list)
    get_val_fn: Any = None

    # for cross validation
    cv_fold: int = None
    depth: int = 0

    def __call__(self):
        if self.cv_fold:
            return cross_val(self.get_train_anno(), self.cv_fold, self.depth)
        else:
            return (
                None if self.get_train_fn is None else self.get_train_anno(),
                None if self.get_val_fn is None else self.get_val_anno())

    def get_train_anno(self):
        return self.get_train_fn(self.target, self.train_dirs)

    def get_val_anno(self):
        return self.get_val_fn(self.target, self.val_dirs)



from typing import List
import re
from pathlib import Path


class CaseDirect:
    def __init__(self, image_dir: str, label_dir: str):
        self.image_dir = image_dir
        self.label_dir = label_dir

    def __call__(self, root: str, *args) -> List[List[Path]]:
        return self.read_case_direct(root)

    def read_case_direct(self, root: str, *args) -> List[List[Path]]:
        """
        - root
            - image_dir
            - label_dir
        """
        annotations = list()
        c_label = Path(root) / self.label_dir
        c_img = Path(root) / self.image_dir
        labels = sorted(self._get_img_files(c_label))
        imgs = [next(c_img.glob(f"{label.stem}.*")) for label in labels]
        annotations = list(zip(imgs, labels))
        return annotations

    def _get_img_files(self, p_dir: Path) -> List[Path]:
        ImageEx = "jpg|jpeg|png|gif|bmp"
        img_files = [
            p for p in p_dir.glob('*')
            if re.search(rf'.*\.({ImageEx})', str(p))
        ]
        return img_files


class Cases(CaseDirect):
    def __init__(self, image_dir: str, label_dir: str):
        super().__init__(image_dir, label_dir)

    def __call__(self, root: str, target_dirs: List[str]) -> List[List[Path]]:
        """
        caseごとにフォルダが作成されている場合
        - root
            - case_name
                - image_dir
                - label_dir
        target_dirs: [case_name1, case_name2, ...]
        """
        annos = list()
        for case_name in target_dirs:
            case_dir = Path(root) / case_name
            annos += self.read_case_direct(
                str(case_dir), self.image_dir, self.label_dir)
        return annos


class CaseGroups(CaseDirect):
    def __init__(self, image_dir: str, label_dir: str):
        super().__init__(image_dir, label_dir)

    def __call__(self, root: str, group_dirs: List[str]) -> List[List[Path]]:
        """
        caseごとのフォルダをさらにグループでまとめている場合
        - root
            - group_name
                - case_name
                    - image_dir
                    - label_dir
        group_dirs: [group_name1, group_name2, ...]
        """
        annos = list()
        for group_name in group_dirs:
            group_dir = Path(root) / group_name
            for case_dir in group_dir.iterdir():
                annos += self.read_case_direct(
                    str(case_dir), self.image_dir, self.label_dir)
        return annos



import torch
import os
from .logger import Logger
from .metrics import SegMetrics
from . import utils
import dataclasses


@dataclasses.dataclass
class GetOptimization:
    train_data: torch.utils.data.Dataset
    val_data: torch.utils.data.Dataset
    batch_size: int
    epochs: int
    lr: float
    gpus: str
    optimizer_cls: torch.optim.Optimizer
    model: torch.nn.Module
    loss_func: torch.nn.Module
    result_dir: str = 'result'
    port: str = '12346'

    def __post_init__(self):
        self.logger = Logger(self.result_dir)
        self.world_size = len(self.gpus.split(","))
        self.is_distributed = self.world_size > 1

    def __call__(self):
        torch.multiprocessing.spawn(
            self.fit, args=(), nprocs=self.world_size, join=True)

    def fit(self, rank):
        self.set_env(rank)
        self.set_params(rank)
        sampler, train_loader, valid_loader = self.make_data_loader()
        for epoch in range(self.epochs):
            self.train(train_loader, rank, sampler, epoch)
            self.validation(valid_loader, rank)
        self.cleanup()

    def set_params(self, rank):
        self.model = self.model.to(rank)
        self.model_without_ddp = self.model
        if self.is_distributed:
            self.model = torch.nn.SyncBatchNorm.convert_sync_batchnorm(
                self.model)
            self.model = torch.nn.parallel.DistributedDataParallel(
                self.model, device_ids=[rank], find_unused_parameters=True)
            self.model_without_ddp = self.model.module
        self.set_optimizer()
        self.set_scheduler()

    def set_optimizer(self):
        self.optimizer = self.optimizer_cls(
            [dict(params=self.model.parameters(), lr=self.lr)])

    def set_scheduler(self):
        self.scheduler = torch.optim.lr_scheduler.LambdaLR(
            self.optimizer, lr_lambda=self.scheduler_func)

    def make_data_loader(self):
        train_sampler = torch.utils.data.distributed.DistributedSampler(
            self.train_data) if self.is_distributed else None
        train_loader = torch.utils.data.DataLoader(
            self.train_data, batch_size=self.batch_size, pin_memory=True,
            shuffle=(train_sampler is None), sampler=train_sampler)
        valid_sampler = torch.utils.data.distributed.DistributedSampler(
            self.val_data) if self.is_distributed else None
        valid_loader = torch.utils.data.DataLoader(
            self.val_data, batch_size=self.batch_size, pin_memory=True,
            shuffle=False, sampler=valid_sampler)
        return train_sampler, train_loader, valid_loader

    def train(self, train_loader, rank, sampler, epoch):
        if self.is_distributed:
            sampler.set_epoch(epoch)
        print(f"\ntrain step, epoch: {epoch + 1}/{self.epochs}")
        self.logger.set_progbar(len(train_loader))
        self.model.train()
        for inputs, labels in train_loader:
            outputs = self.model(inputs.to(rank))
            loss = self.loss_func(outputs, labels.to(rank))
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()
            self.logger(
                    loss.item(),
                    lr=[group['lr'] for group in self.optimizer.param_groups])
        if rank == 0:
            torch.save(
                self.model_without_ddp.state_dict(),
                f'{self.result_dir}/last.pth')

    def validation(self, valid_loader, rank):
        print("\nvalidation step")
        self.logger.set_progbar(len(valid_loader))
        self.model.eval()
        metrics = SegMetrics()
        with torch.no_grad():
            for inputs, labels in valid_loader:
                outputs = self.model(inputs.to(rank))
                loss = self.loss_func(outputs, labels.to(rank))
                confusion = metrics.calc_confusion(outputs, labels.to(rank))
                if self.is_distributed:
                    torch.distributed.all_reduce(confusion)
                dice = metrics.compute_metric(confusion, metrics.dice)
                self.logger(loss.item(), dice=dice.item())
        self.logger.update_metrics()
        self.update_scheduler()

    def update_scheduler(self):
        self.scheduler.step()

    def set_env(self, rank):
        print(f"rank: {rank}")
        os.environ['CUDA_VISIBLE_DEVICES'] = self.gpus
        os.environ['MASTER_ADDR'] = 'localhost'
        os.environ['MASTER_PORT'] = self.port
        torch.distributed.init_process_group(
            "gloo", rank=rank, world_size=self.world_size)
        utils.setup_for_distributed(rank == 0)

    def cleanup(self):
        torch.distributed.destroy_process_group()

    @staticmethod
    def scheduler_func(epoch):
        return 0.95 ** (epoch-10) if epoch > 10 else 1



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['farmer_pytorch',
 'farmer_pytorch.logger',
 'farmer_pytorch.metrics',
 'farmer_pytorch.readers']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=9.1.0,<10.0.0',
 'matplotlib>=3.5.1,<4.0.0',
 'numpy>=1.22.3,<2.0.0',
 'torch>=1.11.0,<2.0.0']

setup_kwargs = {
    'name': 'farmer-pytorch',
    'version': '0.3.0',
    'description': 'deep learning tools: easy to run, easy to customize',
    'long_description': '# Pytorch segmentation\n\n## installation\n```\npip install farmer-pytorch\n```\n\n\n## Quick start\n[quick start](https://github.com/aiorhiroki/fmp_dvc)',
    'author': 'aiorhiroki',
    'author_email': '1234defgsigeru@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/aiorhiroki/farmer.pytorch',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import matplotlib.pyplot as plt
import json
from pathlib import Path
import time
from .progress_bar import ProgressBar


class Logger:
    def __init__(self, result_dir):
        Path(result_dir).mkdir(exist_ok=True, parents=True)
        self.result_dir = result_dir
        self.dice_history = []
        self.start_time = time.time()

    def set_progbar(self, nb_iters):
        self.prog_bar = ProgressBar(nb_iters)

    def __call__(self, loss, dice=None, lr=None):
        self.prog_bar.print_prog_bar(loss, dice, lr)

    def get_latest_metrics(self):
        return self.prog_bar.get_latest_metrics()

    def update_metrics(self):
        self.dice_history += [self.get_latest_metrics()]
        self._plot_logs()
        self._save_metric()

    def _plot_logs(self):
        plt.plot(self.dice_history)
        plt.savefig(f"{self.result_dir}/dice.png")
        plt.close()

    def _save_metric(self):
        took_time = time.time() - self.start_time
        scores = dict(dice=self.dice_history[-1], time=int(took_time))
        with open(f"{self.result_dir}/scores.json", "w") as fw:
            json.dump(scores, fw)

        history_dict = {}
        history_dict['dice'] = []
        for epoch, history in enumerate(self.dice_history):
            history_dict['dice'] += [{'epoch': epoch, 'dice': history}]
        with open(f"{self.result_dir}/dice.json", "w") as fw:
            json.dump(history_dict, fw, indent=4)
